/***************************************************************************
 *   Copyright (C) 2006, IBM                                               *
 *                                                                         *
 *   Maintained By:                                                        *
 *   Eric Munson and Brad Peters                                           *
 *   munsone@us.ibm.com, bpeters@us.ibm.com                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the Lesser GNU General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or at your option) any later version.                        *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the Lesser GNU General Public      *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <libvpd-1/vpddbenv.hpp>
#include <libvpd-1/logger.hpp>
#include <libvpd-1/debug.hpp>

#include <string.h>

static const int KEY_BUFFER_SIZE = 256;

namespace lsvpd
{
	int compareKeys( Db* db, const Dbt* key1, const Dbt* key2 )
	{
		string k1( (char*)( key1->get_data( ) ) );
		string k2( (char*)( key2->get_data( ) ) );
		return k1.compare( k2 );
	}

	/*
	 * The copy constructor is marked private but these things need to be
	 * filled out to compile correctly.
	 */
	 VpdDbEnv::VpdDbEnv( const VpdDbEnv& copyMe ):
	 	mCFlags( copyMe.mCFlags ), mDbFileName( copyMe.mDbFileName ),
	 	mEnvDir( copyMe.mEnvDir ), mpVpdDb( NULL ), mVpdEnv( 0 )
	 	{ }

	VpdDbEnv::VpdDbEnv( const string& envDir, const string& dbFileName,
			u_int32_t cFlags ) :
				mDbFileName( dbFileName ),
				mEnvDir( envDir ),
				mCFlags( cFlags ),
				mpVpdDb( NULL ),
				mVpdEnv( 0 )
	{
		int rc;
		
		try
		{
			/*
			 * Because we are using a DB environment to enable mutil-process
			 * access to the VPD db, we want the db to be used in shared
			 * memory and for the BDB library to provide transactions and
			 * locking so we are forcing the DB_INIT_MPOOL create flag.
			 */
			Logger l;

			rc = mVpdEnv.open( mEnvDir.c_str( ), 
				mCFlags | DB_INIT_MPOOL, 0 );
			if (rc) {
				l.log("Berkeley DB Open Failed - ENV", LOG_ERR);
			}
			mpVpdDb = new Db( &mVpdEnv, 0 );
			mpVpdDb->set_bt_compare( compareKeys );
			rc = mpVpdDb->open( NULL, mDbFileName.c_str( ), NULL, DB_BTREE,
				       mCFlags, 0 );
			if (rc) {
				l.log("Berkeley DB Open Failed.", LOG_ERR);
			}
		}
		catch( DbException& e )
		{
			/*
			 * Because there isn't much that we can do if something in this try
			 * block fails we will only log that an exception occured and then
			 * rethrow it.
			 */
			string message( "VpdDbEnv.VpdDbEnv( " );
			message += envDir;
			message += string( ", " );
			message += dbFileName;
			message += string( " ): " );
			message += string(
				"Must have read permission of /var/lib/lsvpd.  Details: ");
			VpdException ve( message + string( e.what( ) ) );
			Logger l;
			l.log( string( ve.what( ) ), LOG_ERR );

			if( mpVpdDb != NULL )
			{
				delete mpVpdDb;
			}

			throw ve;
		}
	}

	VpdDbEnv::~VpdDbEnv()
	{
		Logger l;
		string msg;
		int rc;

		try
		{
			if( mpVpdDb != NULL )
			{
				rc = mpVpdDb->close( 0 );
				if (rc) {
					Logger l;
					l.log("Berkely DB cleanup failed.", LOG_ERR );
				}
				delete mpVpdDb;
				mpVpdDb = NULL;
			}
			
			rc = mVpdEnv.close( 0 );
			if (rc) {
				Logger l;
				l.log( string("Berkely DB shutdown failed - ENV."), LOG_ERR );
			}
		}
		catch( DbException &e )
		{
			VpdException ve( string( "VpdDbEnv.~VpdDbEnv( ): " ) +
				string( e.what( ) ) );
			Logger l;
			l.log( string( ve.what( ) ), LOG_ERR );

			if( mpVpdDb != NULL )
			{
				delete mpVpdDb;
			}

			throw ve;
		}
	}

	Component* VpdDbEnv::fetch( const string& deviceID )
	{
		Component* ret;
		char buffer[ KEY_BUFFER_SIZE ];

		const char* str = deviceID.c_str( );
		int length = deviceID.length( );

		memset( buffer, '\0', KEY_BUFFER_SIZE );
		memcpy( buffer, str, length );
		Dbt key( buffer, KEY_BUFFER_SIZE );
		Dbt data;

		mpVpdDb->get( NULL, &key, &data, 0 );

		ret = new Component( data.get_data( ) );
		return ret;
	}

	/*
	 * Fetches the system root (the base of the component tree that
	 * stores all the system-wide vpd.
	 */
	System* VpdDbEnv::fetch( )
	{
		System* ret;
		char buffer[ KEY_BUFFER_SIZE ];

		const char* str = System::ID.c_str( );
		int length = System::ID.length( );

		memset( buffer, '\0', KEY_BUFFER_SIZE );

		memcpy( buffer, str, length );
		Dbt key( buffer, KEY_BUFFER_SIZE );
		Dbt data;

		mpVpdDb->get( NULL, &key, &data, 0 );

		ret = new System( data.get_data( ) );
		return ret;
	}

	bool VpdDbEnv::store( Component *storeMe )
	{
		void * buffer = NULL;
		unsigned int dataSize;
		int ret;
		char name [ KEY_BUFFER_SIZE ];
		memset( name, '\0', KEY_BUFFER_SIZE );
		strncpy( name, storeMe->getID( ).c_str( ), KEY_BUFFER_SIZE );
		Dbt key( name, KEY_BUFFER_SIZE );
		Dbt data;
		dataSize = storeMe->pack( &buffer );
		data = Dbt( buffer, dataSize );
		ret = mpVpdDb->put( NULL, &key, &data, 0 );
		if( ret != 0 )
		{
			if( buffer != NULL )
				delete [] (char*)buffer;

			string message(
				"VpdDbEnb.store: BDB error while storing device " );
			message += storeMe->getID( );
			VpdException ve( message );
			throw ve;
		}
		if( buffer != NULL )
			delete [] (char*)buffer;
		return true;
	}

	bool VpdDbEnv::store( System *storeMe )
	{
		void * buffer = NULL;
		unsigned int dataSize;
		int ret;
		char name [ KEY_BUFFER_SIZE ];

		memset( name, '\0', KEY_BUFFER_SIZE );
		strncpy( name, storeMe->getID( ).c_str( ), KEY_BUFFER_SIZE );
		Dbt key( name, KEY_BUFFER_SIZE );
		Dbt data;
		dataSize = storeMe->pack( &buffer );
		data = Dbt( buffer, dataSize );
		ret = mpVpdDb->put( NULL, &key, &data, 0 );
		if( ret != 0 )
		{
			if( buffer != NULL )
				delete [] (char*)buffer;

			string message(
				"VpdDbEnb.store: BDB error while storing device " );
			message += storeMe->getID( );
			VpdException ve( message );
			throw ve;
		}
		if( buffer != NULL )
			delete [] (char*)buffer;
		return true;
	}

	bool VpdDbEnv::remove( const string& deviceID )
	{
		char buffer[ KEY_BUFFER_SIZE ];
		bool ret = true;

		memset( buffer, '\0', KEY_BUFFER_SIZE );
		memcpy( buffer, deviceID.c_str( ), deviceID.length( ) );

		Dbt key( (void*)buffer, KEY_BUFFER_SIZE );

		if( mpVpdDb->del( NULL, &key, 0 ) != 0 )
		{
			ret = false;
		}
		return ret;
	}

	vector<string> VpdDbEnv::getKeys( )
	{
		vector<string> ret;
		Dbc * pCursor = NULL;
		Dbt key, data;

		try
		{
			mpVpdDb->cursor( NULL, &pCursor, 0 );

			if( pCursor == NULL )
			{
				return ret;
			}

			while( pCursor->get( &key, &data, DB_NEXT ) == 0 )
			{
				ret.push_back( string( (char*)key.get_data( ) ) );
			}

			pCursor->close( );
		}
		catch( DbException& e )
		{
			Logger l;
			string message( "VpdDbEnv.getKeys: DbException caught: " );
			message += e.what( );
			message += " returning the keys collected so far.";
			l.log( message, LOG_WARNING );
			return ret;
		}
		return ret;
	}

	int VpdDbEnv::countEntries( )
	{
		int ret = 0;
		Dbc * pCursor = NULL;
		Dbt key, data;
		try
		{
			mpVpdDb->cursor( NULL, &pCursor, 0 );

			if( pCursor == NULL )
			{
				return ret;
			}

			while( pCursor->get( &key, &data, DB_NEXT ) == 0 )
			{
				ret++;
			}

			pCursor->close( );
		}
		catch( DbException& e )
		{
			Logger l;
			string message( "VpdDbEnv.getKeys: DbException caught: " );
			message += e.what( );
			message += " returning the number of entries collected so far.";
			l.log( message, LOG_WARNING );
			return ret;
		}
		return ret;
	}
}
