/***************************************************************************
 *   Copyright (C) 2006, IBM                                               *
 *                                                                         *
 *   Maintained by:                                                        *
 *   Eric Munson and Brad Peters                                           *
 *   munsone@us.ibm.com, bpeters@us.ibm.com                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *********************L******************************************************/

#ifdef TRACE_ON
	#include <libvpd-1/debug.hpp>
#endif
#ifdef HAVE_CONFIG_H
#include <libvpd-1/config.h>
#endif

#define _XOPEN_SOURCE 500 // For pread

#include <libvpd-1/sysfstreecollector.hpp>
#include <libvpd-1/helper_functions.hpp>
#include <libvpd-1/debug.hpp>
#include <libvpd-1/logger.hpp>
#include <libvpd-1/lsvpd.hpp>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/sysmacros.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <linux/types.h>
#include <linux/sockios.h>
#include <linux/ethtool.h>
#include <linux/hdreg.h>
#include <net/if_arp.h>
#include <net/if.h>

#include <deque>
#include <fstream>
#include <sstream>
#include <iomanip>

extern int errno;

using namespace std;

namespace lsvpd
{
	SysFSTreeCollector::SysFSTreeCollector( )
	{
		ifstream id;
		mPciTable = NULL;
		mUsbTable = NULL;

		id.open( DeviceLookup::PCI_ID_FILE.c_str( ), ios::in );
		if( id )
		{
			mPciTable = new DeviceLookup( id );
			id.close( );
		}
		else
		{
			Logger logger;
			logger.log(
				"pci.ids file not found, continuing but there will be a lot of missing information",
				LOG_ERR );
		}

		id.open( DeviceLookup::USB_ID_FILE.c_str( ), ios::in );
		if( id )
		{
			mUsbTable = new DeviceLookup( id );
			id.close( );
		}
		else
		{
			Logger logger;
			logger.log(
				"usb.ids file not found, continuing but there will be a lot of missing information",
				LOG_ERR );
		}
	}

	SysFSTreeCollector::~SysFSTreeCollector( )
	{
		if( mPciTable != NULL )
			delete mPciTable;

		if( mUsbTable != NULL )
			delete mUsbTable;
	}

	void SysFSTreeCollector::scsiGetHTBL(Component *fillMe)
	{
		string tmp = fillMe->sysFsNode.getValue();
		int loc, ploc;

		if ( (loc = tmp.rfind("/",tmp.length())) != string::npos ) {
			tmp = tmp.substr(loc + 1, tmp.length());

			if ( (loc = tmp.find(":", 0)) != string::npos)
				fillMe->addDeviceSpecific("XH", "SCSI Host", tmp.substr(0, loc), 60);

			ploc = loc + 1;
			if ( (loc = tmp.find(":", ploc)) != string::npos)
				fillMe->addDeviceSpecific("XB", "SCSI Bus",
													tmp.substr(ploc, loc - ploc), 60);
			ploc = loc + 1;
			if ( (loc = tmp.find(":", ploc)) != string::npos)
				fillMe->addDeviceSpecific("XT", "SCSI Target",
											tmp.substr(ploc, loc - ploc), 60);
			ploc = loc + 1;
			fillMe->addDeviceSpecific("XL", "SCSI Lun",
											tmp.substr(ploc, tmp.length() - ploc), 60);

		}
	}
	void SysFSTreeCollector::ideGetHTBL(Component *fillMe)
	{
		string tmp = fillMe->sysFsNode.getValue();
		int loc;

		if ( (loc = tmp.rfind("/",tmp.length())) != string::npos )
			fillMe->addDeviceSpecific("XI", "IDE Device Identifier",
					tmp.substr(loc + 1, tmp.length()), 60);

	}

	void SysFSTreeCollector::usbGetHTBL(Component *fillMe)
	{
//		coutd << "USB: Cur Dev Node = " <<  fillMe->sysFsNode.getValue() << endl;
	}

	/**
	 * @brief: Collects vpd for devices specified by bus.  This is necessary
	 * 	since the bus of a device can determine the methods used to
	 * 	collect vpd.
	 * @arg bus: The devices bus, as seen at /sys/bus/XXX
	 * @arg fillMe: Device component to be filled
	 */
	Component *SysFSTreeCollector::fillByBus(const string& bus,
		Component* fillMe)
	{
		// Component is a scsi device - fill accordingly
		if(bus  == "scsi" )
		{
			fillMe->mManufacturer.setValue( getAttrValue(
				fillMe->sysFsNode.getValue() , "vendor" ), 50, __FILE__, __LINE__);

			fillMe->mDescription.setValue( getAttrValue(
				fillMe->sysFsNode.getValue() , "device" ), 50, __FILE__, __LINE__);

			fillSCSIComponent(fillMe);  //Fills SH, ST, SB, SL

		}
		else if( bus == "usb" )
		{
			fillUSBDev( fillMe, fillMe->sysFsNode.getValue() );
			/*
			 * Since usb storage devices are treated as SCSI devices,
			 * we can get useful information from sg_utils
			 */
			fillSCSIComponent(fillMe);
			usbGetHTBL(fillMe);  //Get identifiers for device
		}
		else if( bus == "ide" )
		{
			fillIDEDev( fillMe );
			ideGetHTBL(fillMe);  //Get identifiers for device
		}
		else
		{
			fillPCIDev( fillMe, fillMe->sysFsNode.getValue() );
		}

		return fillMe;
	}

	/**
	 * @brief: Fill a device component by class.
	 */
	Component *SysFSTreeCollector::fillByDevClass(const string& devClass,
		Component* fillMe)
	{
		string classLink;
		Logger l;
		string msg;
		
		/* Here we need to do any of the class specific filling necessary.
		 * the device class was populated above when the Kernel name was set.
		 */

		if( devClass == "net" )
		{
			classLink = getClassLink( fillMe );
			
			if (classLink.length() > 0)
				fillNetClass( fillMe,  classLink);
		}

		return fillMe;
	}

	/**
	 * This is the second biggest job for any collector.  Now that we have
	 * identified all the unique devices available on the system, the
	 * collector will attempt to fill out all the device information
	 * available from its data source for the device passed in as 'fillMe'.
	 */
	Component* SysFSTreeCollector::fillComponent( Component* fillMe )
	{
		initComponent(fillMe);

		if( fillMe->sysFsNode.getValue() == "" )
		{
			return fillMe;
		}

		fillByBus(fillMe->devBus.getValue(), fillMe);

		fillByDevClass(fillMe->mDevClass.getValue(), fillMe);

		fillFirmware( fillMe );

		ostringstream os;

		if( fillMe->mpParent != NULL &&
			fillMe->mpParent->mPhysicalLocation.dataValue != "" )
		{
			if( fillMe->mpParent->mPhysicalLocation.dataValue[ 0 ] != 'U' )
				os << fillMe->mpParent->mPhysicalLocation.dataValue << "/";
			else if( fillMe->mpParent->devBusAddr.dataValue != "" )
				os << fillMe->mpParent->devBusAddr.dataValue << "/";
		}

		os << fillMe->devBusAddr.dataValue;

		fillMe->mPhysicalLocation.setValue( os.str( ), 50 , __FILE__, __LINE__);

		if( fillMe->mDescription.dataValue == "" &&
			fillMe->mModel.dataValue != "" )
			fillMe->mDescription.setValue( fillMe->mModel.dataValue, 1 , __FILE__, __LINE__);

		return fillMe;
	}

	void SysFSTreeCollector::initComponent( Component* in )
	{
	}

	/**
	 * hotplugDiscoverComponent(const string& , Component* )
	 *
	 * @brief: used for initial high-level fill of HOTPLUG devices,
	 * making them suitable for passing in to fillComponent()
	 * @var deviceID: The sysfs node of the new device.  Note: Caller is
	 * 		responsible for ensuring this is an absolute path, rather than
	 * 		relative, as provided by the hotplug script.
	 * @var fillMe: Component to be filled
	 * @caller
	 *   Note: This in theory will be called from gatherer, whenever it becomes
	 * 		aware of a hotplug ACTION="ADD"
	 */
	void SysFSTreeCollector::hotplugDiscoverComponent(const string& sysFsNode,
		Component * fillMe)
	{
		Logger l;
		int i;
		string tmp;
		int locBeg, locEnd;

		l.log( "hotplugDiscoverComponent() : Adding '" + sysFsNode + "'", LOG_NOTICE );

		//Set path to dev
		fillMe->sysFsNode.setValue(sysFsNode, INIT_PREF_LEVEL, __FILE__, __LINE__);
		fillMe->idNode.setValue(sysFsNode, INIT_PREF_LEVEL, __FILE__, __LINE__);
		fillMe->mParent.setValue( "/sys/bus", INIT_PREF_LEVEL, __FILE__, __LINE__ );

		//Set what we know of name
		tmp = sysFsNode;
		HelperFunctions::fs_fixPath(tmp);
		i = tmp.length();
		while (tmp[i--] != '/') {}
		tmp = tmp.substr(i, tmp.length());
		fillMe->devSysName.setValue(tmp, INIT_PREF_LEVEL - 1, __FILE__, __LINE__);

		// Find bus name in sysfs path
		locBeg = 1;  //Start at 1 to skip the first "/" or "%c" issue
		for (int i = 0; i < 2; i++)
			locBeg = fillMe->sysFsNode.getValue().find("/", locBeg+1);
		locEnd = fillMe->sysFsNode.getValue().find("/", locBeg+1);
		fillMe->devBus.setValue(fillMe->sysFsNode.getValue().substr(
			locBeg + 1, locEnd - locBeg -1), INIT_PREF_LEVEL, __FILE__, __LINE__);

		// Find bus address in sysfs path
		locBeg = 1;  //Start at 1 to skip the first "/" or "%c" issue
		for (int i = 0; i < 4; i++)
			locBeg = fillMe->sysFsNode.getValue().find("/", locBeg+1);
		locEnd = fillMe->sysFsNode.getValue().find("/", locBeg+1);
		fillMe->devBusAddr.setValue(fillMe->sysFsNode.getValue().substr(
			locBeg + 1, locEnd - locBeg -1), INIT_PREF_LEVEL,
			__FILE__, __LINE__);

		char buf[ 2048 ] = { 0 };
		if( readlink( sysFsNode.c_str( ), buf, 2047 ) > 0 )
		{
			char* p = strdup( sysFsNode.c_str( ) );
			string linkTarget = HelperFunctions::getAbsolutePath( buf, p );
			free( p );
			string parentDev = linkTarget.substr( 0, linkTarget.rfind( '/' ) );
			struct stat info;
			if( stat( parentDev.c_str( ), &info ) == 0 )
			{
				ostringstream base;
				base << sysFsNode.substr( 0, sysFsNode.rfind( '/' ) );
				base << "/" << parentDev.substr( parentDev.rfind( '/' ) + 1 );
				fillMe->mParent.setValue( base.str( ), 60, __FILE__, __LINE__ );
			}
		}

		// Pointer from sysfs -> device-tree node
		tmp = getDevTreePath(fillMe->sysFsNode.getValue());
		if (tmp.length() > 0) {
			fillMe->deviceTreeNode.setValue("/proc/device-tree" + tmp,
				INIT_PREF_LEVEL - 1, __FILE__, __LINE__);
		}

	}

	/**
	 * For SysFSTreeCollector, init will test to see if path_t is,
	 *  available.  If it is not, we cannot use this collector, as the
	 *  /sys file system is lacking.
	 * Note: This will substantially and negatively impact device
	 * 	data collection.
	 */
	bool SysFSTreeCollector::setup(const string path_t )
	{
		struct stat statbuf;
		bool ret;

		if (stat(path_t.c_str(), &statbuf) != 0)
			ret = false;
		else {
			rootDir = path_t;
			fsw = FSWalk(rootDir);
			ret = true;
		}

		return ret;
	}

	bool SysFSTreeCollector::init()
	{
		return setup("/sys/bus");
	}

	/**
	 * Walkes a vector of component ptr's, looking for a particular one
	 * @return the specified component, or NULL on failure
	 */
	Component *SysFSTreeCollector::findComponent(const vector<Component*> devs,
		string sysPath )
	{
		for (int i = 0; i < devs.size(); i++)
			if (devs[i]->sysFsLinkTarget.getValue() == sysPath) {
				return devs[i];
			}

		return NULL;
	}

	/* Creates a vector of components, representing all devices this
	 *		collector is aware of (ie - those discovered in the /sys tree),
	 *		and back-walks tree to determine inheritance
	 * @arg devs: An empty vector, which will be filled with one Component
	 * 	for each device on the system.
	 */
	vector<Component*> SysFSTreeCollector::getComponents(
		vector<Component*>& devs )
	{
		Component *child, *parent;
		string sysPath, str;
		int i;

		devs = getComponentsVector(devs);

		for (i = (devs.size() - 1); i >= 0; i--) {
			child = devs[i];
			sysPath = child->sysFsLinkTarget.getValue();

			/*
			 * Need to set up sysFsClassNode constituents - ie  mDevClass and
			 * mAIXName
			 */
			setKernelName( child, child->sysFsNode.getValue());

			parent = NULL;
			while (HelperFunctions::dropDir(sysPath)) {
				parent = findComponent(devs, sysPath);

				if (NULL != parent) {
					child->mParent.setValue(parent->idNode.getValue(),
						INIT_PREF_LEVEL + 1, __FILE__, __LINE__);
					parent->addChild(child->idNode.getValue());
					break;
				}
			}

			if (child->devBus.getValue() == "scsi")
				scsiGetHTBL(child);  //Get Host:Target:Bus:Lun data for device

		}

		/*
		 * This case indicates that we are running on a older kernel that does
		 * not maintain the links from /sys/pci/XXX/devices/XXX into the
		 * corresponding /sys/class/XXX/XXX directory so we cannot use the
		 * above method for finding kernel names, we must fall back on
		 * the method found in scripts/scan.d/[03adapter|07device].
		 */

		/* If this kernel does not provide links to the class dir,
		 * we will try a backup approach to filling AX.
		 *
		 * This function now works to walk the entire tree of directories just once,
		 * while walking the device list N times.  This should substantially
		 * cut down on processing time as device number << directories */

		readClassEntries( devs );

	return devs;
	}

	/**
	 * Retrieves the path to the device-tree representation of sysFS device
	 * at sysPath, if it exists.  Otherwise, returns string("")
	 */
	string SysFSTreeCollector::getDevTreePath(string sysPath)
	{
		struct stat astats;
		const char *buf;
		char buf2[512];
		FILE *fi;
		int i;

        HelperFunctions::fs_fixPath(sysPath);
        sysPath += "/devspec";

        buf = sysPath.c_str();
        if ((lstat(buf, &astats)) != 0) {
            return string("");
        }

		// Read Results
		fi = fopen(buf , "r");
		if (fi != NULL) {
			fgets(buf2, 512, fi);
			i = 0;
			while (buf2[i] != '\0') {
				if (buf2[i] == '\n')
					buf2[i] = '\0';
				i++;
			}
			fclose(fi);
		}

		// cleanup
		return string(buf2);
	}

	/**
	 * getComponents( vector<Component*>& devs )
	 * @Note: This scales as O(n^4)... ugly.  Is there a better way?
	 *
	 * @brief Manually walks the sysfs tree, ID's all devices, adds path and
	 * other relevant data to the components vector.  Must call fillComponent
	 * on these to obtain full details
	 *
	 * @var devs: Empty vector which will be filled with all devices this
	 * collector knows how to elucidate
	 */
	vector<Component*> SysFSTreeCollector::getComponentsVector(
		vector<Component*>& devs )
	{
		vector<string> curList;
		vector<string> devNodeList, subDiscoveredDevs;
		deque<string> fullList;		// All devices get pushed into here
		int fullListFrontNode = 0;
		string curPath, tmpDirName, linkName, filePath, devPath;
		Component *tmp;
		char *buf;
		FSWalk fsw = FSWalk();
		bool isDev, dev, found;
		char linkTarget[512];
		int locBeg, locEnd;
		if (rootDir.length()== 0)
			rootDir = string("/sys/bus");

		HelperFunctions::fs_fixPath(rootDir);
		fullList.push_back(rootDir); //Start at top of device tree

		while (fullList.size() > 0 )
		{
			curPath = fullList.front();
			fullList.pop_front();

			fsw.fs_getDirContents(curPath, 'd', curList);
			/* curList is simply the entire listing of directory entries.
			 * We walk across each, and determine if they are links, files, or
			 * dirs
			 */
			while (curList.size() > 0)
			{
				tmpDirName = curList.back();
				dev = false;

				if (tmpDirName == "devices")
				{
					tmpDirName = curPath + "/" + tmpDirName;
					/*
					 * Get info on each device found in this /devices dir
					 */
					fsw.fs_getDirContents(tmpDirName, '*', devNodeList);
					while (devNodeList.size() > 0) {
						int len;
						string absTargetPath;

						linkName = devNodeList.back();
						devNodeList.pop_back();
						filePath = tmpDirName + "/" + linkName;

						// Create new component
						tmp = new Component();

						// If link - resolve to absolute path
						if (FSWalk::fs_isLink(filePath))
						{
							buf = strdup(filePath.c_str());
							len = readlink(buf, linkTarget, sizeof(
								linkTarget));
							linkTarget[len] = '\0';
							absTargetPath = HelperFunctions::getAbsolutePath(linkTarget,
								buf);
							tmp->sysFsLinkTarget.setValue(absTargetPath,
								INIT_PREF_LEVEL + 1, __FILE__, __LINE__);
						}

						/* -------- Partially fill component with
						 * immediately necessary/available fields ------*/
						//Set path to dev
						tmp->sysFsNode.setValue(filePath, INIT_PREF_LEVEL
														, __FILE__, __LINE__);
						tmp->idNode.setValue(filePath, INIT_PREF_LEVEL
														, __FILE__, __LINE__);

						// get node_type: /sys/bus/XXX  OR sys/bus/XXX

						// Find bus name in sysfs path
						//START AT 1 TO SKIP THE FIRST "/" OR "X" ISSUE
						locBeg = 1;
						for (int i = 0; i < 2; i++)
							locBeg = tmp->sysFsNode.getValue().find("/",
								locBeg+1);
						locEnd = tmp->sysFsNode.getValue().find("/",
							locBeg+1);
						tmp->devBus.setValue(tmp->sysFsNode.getValue().
							substr(locBeg + 1, locEnd - locBeg -1),
							INIT_PREF_LEVEL, __FILE__, __LINE__);

						// Find bus address in sysfs path
						string sysNode = tmp->sysFsNode.getValue();
						int lastSlash = sysNode.rfind("/", sysNode.length()) + 1;
						tmp->devBusAddr.setValue(sysNode.substr(lastSlash,
														sysNode.length() - lastSlash),
											INIT_PREF_LEVEL, __FILE__, __LINE__);
						//Set what we know of name
						tmp->devSysName.setValue(linkName,
							INIT_PREF_LEVEL - 1, __FILE__, __LINE__);
						devs.push_back(tmp);

						/*
						 * default parent is top object.
						 * This will be overwritten as more detailed info
						 * becomes available
						 */
						 tmp->mParent.setValue("/sys/bus", INIT_PREF_LEVEL,
						 								__FILE__, __LINE__);

						// Pointer from sysfs -> device-tree node
						devPath = getDevTreePath(tmp->sysFsNode.getValue());
						if (devPath.length() > 0)
						{
							tmp->deviceTreeNode.setValue(
								"/proc/device-tree" + devPath,
								INIT_PREF_LEVEL - 1, __FILE__, __LINE__);
						}
						dev = true;
					}
				} 
				/* the isDir check seems to randomly fail at times, so
				 * instead of checking if we have found a subdir, we check 
				 * to ensure this is not a link.  Since the fs_getDirContents()
				 * call was for type 'd', this returned list can only be dir's
				 * or links that were mistakenly identified as dirs.
				 * we DO NOT want to follow links, or we end up in circular,
				 * infinite loops */
				else if (! FSWalk::fs_isLink(filePath))
				{
					// Push all dirs into fullList for future walking
					fullList.push_back(curPath + "/" + tmpDirName);
				}

				curList.pop_back();
			}
			fullListFrontNode++;
		}

		return devs;
	}

	/* Collects devices on system, then returns how many were found.
	 */
	int SysFSTreeCollector::numDevicesInTree()
	{
		vector<Component*> devs;

		getComponentsVector(devs);
		return devs.size();
	}

	/* Name of this collector */
	string SysFSTreeCollector::myName(void)
	{
		return string("SysFSTreeCollector");
	}

	/**
	 * @brief: Walks thru all links in this directory, looking for one
	 * 	with a ':' in the name.  This one has always been the back
	 * 	link to the /sys/class dir, where more info is gathered.
	 * Note: This may break on occassion, so a better method should be
	 * 	investigated.
	 * @arg: The directory in which the device was found
	 */
	string SysFSTreeCollector::getClassLink( const string& sysDir )
	{
		Logger logger;
		struct dirent* entry;
		DIR* d;
		string link;

		string ret = "";

		d = opendir( sysDir.c_str( ) );
		if( d == NULL ) {
			return ret;
		}

		while( ( entry = readdir( d ) ) != NULL )
		{
			string fname = entry->d_name;
			if( HelperFunctions::countChar( fname, ':' ) == 1 )
			{
				char* f, *p;
				f = strdup( fname.c_str( ) );
				p = strdup( sysDir.c_str( ) );
				link = HelperFunctions::getAbsolutePath( f, p );
				free( f );
				free( p );
				break;
			}
		}
		closedir( d );
		return ret;
	}

	/* Prints class node for device */
	string SysFSTreeCollector::getClassLink( const Component* comp )
	{
		if( comp->mAIXNames.empty( ) )
			return "";

		ostringstream out;
		out << "/sys/class/" << comp->mDevClass.dataValue << "/";
		out << comp->mAIXNames[ 0 ]->dataValue << "/";
		return out.str( );
	}

	/* Attempts a best effort collection of AIX name, of AX
	 * @param sysDir: Dir to start a directory walk, looking for
	 * 		link fitting the know class link format (ie ':' present)
	 * @return bool: Whether the AX name was successfully filled
	 */
	bool SysFSTreeCollector::setKernelName( Component* fillMe,
		const string& sysDir )
	{
		Logger logger;
		struct dirent* entry;
		DIR* d;
		string link;
		bool filled = false;

		d = opendir( sysDir.c_str( ) );
		if( d == NULL )
		{
			if (sysDir.length() > 0) {
				ostringstream os;
				os << "Unable to open requested dir '" << sysDir << "'";
				logger.log( os.str( ), LOG_WARNING );
			}
			return true;
		}

		while( ( entry = readdir( d ) ) != NULL )
		{
			string fname = entry->d_name;
			int idx;
			if( HelperFunctions::countChar( fname, ':' ) == 1 )
			{
				idx = fname.find( ':' );
				idx++;
				if( !HelperFunctions::contains( fillMe->mAIXNames,
					fname.substr( idx ) ) )
				{
					ostringstream dirName;
					dirName << "/sys/class/" << fname.substr( 0, idx - 1 ) <<
						"/" << fname.substr( idx );
					struct stat info;
					if( stat( dirName.str( ).c_str( ), &info ) == 0 )
					{
						filled = true;
						fillMe->addAIXName( fname.substr( idx ), 90 );
						fillMe->mDevClass.setValue(
							fname.substr( 0, idx - 1 ), 90, __FILE__,
							__LINE__ );
					}
					else if( fname.substr( 0, idx - 1 ) == "block" )
					{
						filled = true;
						fillMe->addAIXName( fname.substr( idx ), 90 );
						fillMe->mDevClass.setValue(
							fname.substr( 0, idx - 1 ), 90, __FILE__,
							__LINE__ );
					}
				}
			}
		}
		closedir( d );
		return filled;
	}

	void SysFSTreeCollector::readClassEntries( vector<Component*>& devs )
	{
		struct dirent* entry;
		DIR* d;
		d = opendir( "/sys/class" );

		if( d == NULL ) {
			return;
		}

		while( ( entry = readdir( d ) ) != NULL ) {
			string cls = entry->d_name;
			if (cls != "." && cls != "..")
				readClassDevs( devs, "/sys/class/", cls );
		}

		// Now look in /sys/block
		readClassDevs( devs, "/sys/", "block" );
		closedir( d );
	}

	/* readClassDevs()
	 * @brief: Walk through full dir tree provided, starting at 'base',
	 * 	executing 'readClassDevice()' on each dir
	 */
	void SysFSTreeCollector::readClassDevs( vector<Component*>& devs,
		const string& base, const string& cls )
	{
		struct dirent* entry;
		DIR* d;
		string path = base + cls;

		d = opendir( path.c_str( ) );
		if( d == NULL ) {
			return;
		}

		while( ( entry = readdir( d ) ) != NULL ) {
			string dev = entry->d_name;
			if (dev != "." && dev != "..")
				readClassDevice( devs, base, cls, dev );
		}

		closedir( d );
	}

	/* readClassDevice()
	 * @brief: Look through directory for back links to each devices
	 * 		/sys/bus/* entry.  If match, we can use this dir for
	 * 		additional information.  Ex: AIX names, deviceTreeNode's, etc
	 */
	void SysFSTreeCollector::readClassDevice( vector<Component*>& devs,
		const string& base, const string& cls, const string& dev )
	{
		vector<Component*>::iterator i, last;
		char sym[ 1024 ] = { 0 };
		string path = base + cls + "/" + dev + "/device";

		char* dup = strdup( path.c_str( ) );
		if( readlink( path.c_str( ), sym, 1023 ) < 0 )
		{
			return;
		}

		string abs = HelperFunctions::getAbsolutePath( sym, dup );
		free( dup );

		/* Walk each device, look for match to this link */
		for( i = devs.begin( ), last = devs.end( ); i != last; ++i )
		{
			if ((*i)->sysFsLinkTarget.getValue() == abs )
			{
				if( !HelperFunctions::contains( (*i)->getAIXNames( ), dev ) )
				{
					(*i)->addAIXName( dev, 90 );
					(*i)->mDevClass.setValue( cls, 90, __FILE__, __LINE__ );
				}
				return;
			}
		}
	}

	void SysFSTreeCollector::fillIDEDev( Component* fillMe )
	{
		if( fillMe->mAIXNames.empty( ) ) {
			return;
		}

		ostringstream os;
		string val;
		val = fillMe->mAIXNames[ 0 ]->dataValue;
		os << "/sys/block/" << val << "/device/modalias";
		ifstream in;
		in.open( os.str( ).c_str( ) );
		if( in ) {
			val = "";
			in >> val;
			in.close( );
			if( val.find( "cdrom" ) != string::npos )
			{
				fillMe->mDescription.setValue( "IDE Optical Drive", 80,
					__FILE__, __LINE__);
			}
			else if( val.find( "disk" ) != string::npos )
			{
				fillMe->mDescription.setValue( "IDE Disk Drive", 80,
					__FILE__, __LINE__);
			}
			else if( val.find( "tape" ) != string::npos )
			{
				fillMe->mDescription.setValue( "IDE Tape Drive", 80,
					__FILE__, __LINE__) ;
			}
			else if( val.find( "floppy" ) != string::npos )
			{
				fillMe->mDescription.setValue( "IDE Floppy Drive", 80,
					__FILE__, __LINE__) ;
			}
			else
			{
				fillMe->mDescription.setValue( "IDE Unknown Device", 80,
					__FILE__, __LINE__) ;
			}
		}

		os.str( "" );

		os << "/dev/" << fillMe->mAIXNames[ 0 ]->dataValue;
		struct hd_driveid id;
		memset( &id, 0, sizeof( struct hd_driveid ) );
		int fd = open( os.str( ).c_str( ), O_RDONLY | O_NONBLOCK );
		if( fd < 0 )
		{
			Logger logger;
			string msg = "Failed to open " + os.str( );
			logger.log( msg, LOG_WARNING );
			return;
		}

		if( ioctl( fd, HDIO_GET_IDENTITY, &id ) != 0 )
		{
			Logger logger;
			logger.log( "SysFsTreeCollector.fillIDEDev: ioctl call failed.",
				LOG_WARNING );
			close( fd );
			return;
		}
		close( fd );

		val = (char*)id.model;
		if( val != "" )
		{
			fillMe->mModel.setValue( val, 50, __FILE__, __LINE__ );
		}
		val = (char*)id.fw_rev;
		if( val != "" )
		{
			if( val.find( fillMe->mModel.dataValue ) != string::npos )
			{
				val = val.substr( 0,  val.find( fillMe->mModel.dataValue ) );
			}
			fillMe->mFirmwareVersion.setValue( val, 80, __FILE__, __LINE__);
		}
		val = (char*)id.serial_no;
		if( string( val ) != "" )
		{
			fillMe->mSerialNumber.setValue( val, 80, __FILE__, __LINE__ );
		}
	}

	void SysFSTreeCollector::fillPCIDev( Component* fillMe,
		const string& sysDir )
	{
		string val;
		int manID = UNKNOWN_ID, devID = UNKNOWN_ID;
		int subMan = UNKNOWN_ID, subID = UNKNOWN_ID;

		setKernelName( fillMe, sysDir );

		// Get the ID numbers from sysfs.
		val = getAttrValue( sysDir, "vendor" );
		if( val != "" )
		{
			manID = strtol( val.c_str( ), NULL, 16 );
		}

		val = getAttrValue( sysDir, "device" );
		if( val != "" )
		{
			devID = strtol( val.c_str( ), NULL, 16 );
		}

		val = getAttrValue( sysDir, "subsystem_vendor" );
		if( val != "" )
		{
			subMan = strtol( val.c_str( ), NULL, 16 );
		}

		val = getAttrValue( sysDir, "subsystem_device" );
		if( val != "" )
		{
			subID = strtol( val.c_str( ), NULL, 16 );
		}

		ostringstream os;

		if( mPciTable != NULL )
		{
			// Fill Manufacturer Name
			if( subMan == UNKNOWN_ID )
			{
				if( manID != UNKNOWN_ID )
				{
					val = mPciTable->getName( manID );
					if( val != "" )
						fillMe->mManufacturer.setValue( val, 50, __FILE__, __LINE__ );
					os << hex << setw( 4 ) << setfill( '0' ) << manID;
				}
			}
			else
			{
				val = mPciTable->getName( subMan );
				if( val != "" )
					fillMe->mManufacturer.setValue( val, 50, __FILE__, __LINE__ );
				os << hex << setw( 4 ) << setfill( '0' ) << subMan;
			}

			// Fill Device Model
			if( subID == UNKNOWN_ID || mPciTable->getName( manID, devID,
				subID ) == "Unknown" )
			{
				if( manID != UNKNOWN_ID )
				{
					os << hex << setw( 4 ) << setfill( '0' )<< devID;
					val = mPciTable->getName( manID, devID );
					if( val != "" )
						fillMe->mModel.setValue( val, 80, __FILE__, __LINE__ );
				}
			}
			else
			{
				os << hex << setw( 4 ) << setfill( '0' )<< subID;
				val = mPciTable->getName( manID, devID, subID );
				if( val != "" )
					fillMe->mModel.setValue( val, 80, __FILE__, __LINE__ );
				val = mPciTable->getName( manID, devID );
				if( val != "" )
					fillMe->mDescription.setValue( val, 80, __FILE__, __LINE__ );
			}

			if( os.str( ) != "ffffffff" )
				fillMe->mCDField.setValue( os.str( ), 100, __FILE__, __LINE__ );
		}

		fillPCIDS( fillMe );

		// Read the pci config file for Device Specific (YC)
		os.str( "" );
		os << fillMe->sysFsNode.dataValue << "/config";
		int fd, size;
		char data = 0;
		fd = open( os.str( ).c_str( ), O_RDONLY );
		if( fd < 0 )
		{
			return;
		}

		if( pread( fd, &data, 1, 8 ) < 1 )
		{
			close( fd );
			return;
		}

		close( fd );
		os.str( "" );

		size = (int)data;
		if( size < 0 )
			size *= -1;
		os << dec << (int)size;
		fillMe->addDeviceSpecific( "YC", "Device Specific", os.str( ), 100 );
	}

	void SysFSTreeCollector::fillPCIDS( Component* fillMe )
	{
		string val = fillMe->mDevClass.dataValue;

		if( val == "net" )
			fillMe->mDescription.setValue( "Ethernet PCI Adapter", 30,
						__FILE__, __LINE__ );
		else if( val == "graphics" )
			fillMe->mDescription.setValue( "Display Adapter", 30,
						__FILE__, __LINE__ );
		else if( val == "sound" )
			fillMe->mDescription.setValue( "Multimedia Audio Controller", 30,
						__FILE__, __LINE__ );
		else if( val == "serial" )
			fillMe->mDescription.setValue( "Serial Adapter", 30,
						__FILE__, __LINE__ );
		else if( val == "scsi_device" )
			fillMe->mDescription.setValue( "SCSI Device", 30,
						__FILE__, __LINE__ );
		else if( val == "scsi_disk" )
			fillMe->mDescription.setValue( "SCSI Disk", 30,
						__FILE__, __LINE__ );
		else if( val == "scsi_generic" )
			fillMe->mDescription.setValue( "Generic SCSI Device", 30,
						__FILE__, __LINE__ );
		else if( val == "scsi_host" )
			fillMe->mDescription.setValue( "SCSI Adapter", 30,
						__FILE__, __LINE__ );
		else if( val == "bluetooth" )
			fillMe->mDescription.setValue( "Bluetooth Device", 30,
						__FILE__, __LINE__ );
		else if( val == "pcmcia_socket" )
			fillMe->mDescription.setValue( "PCMCIA Adapter", 30,
						__FILE__, __LINE__ );
		else if( val == "usb" )
			fillMe->mDescription.setValue( "Generic USB Device", 30,
						__FILE__, __LINE__ );
		else if( val == "usb_device" )
			fillMe->mDescription.setValue( "USB Device", 30,
						__FILE__, __LINE__ );
		else if( val == "usb_host" )
			fillMe->mDescription.setValue( "USB Host Controller", 30,
						__FILE__, __LINE__ );
		else if( val == "block" )
			fillMe->mDescription.setValue( "Generic Block Device", 10,
						__FILE__, __LINE__ );
		else if( val == "tty" )
			fillMe->mDescription.setValue( "Serial Device", 20,
						__FILE__, __LINE__ );
	}

	void SysFSTreeCollector::fillUSBDev( Component* fillMe,
		const string& sysDir )
	{
		string val;
		int manID = UNKNOWN_ID, devID = UNKNOWN_ID;

		setKernelName( fillMe, sysDir );

		// Get the ID numbers from sysfs.
		val = getAttrValue( sysDir, "idVendor" );
		if( val != "" )
		{
			manID = strtol( val.c_str( ), NULL, 16 );
		}

		val = getAttrValue( sysDir, "idProduct" );
		if( val != "" )
		{
			devID = strtol( val.c_str( ), NULL, 16 );
		}

		if( mUsbTable != NULL )
		{
			// Fill Manufacturer Name
			if(  manID != UNKNOWN_ID )
			{
				fillMe->mManufacturer.setValue( mUsbTable->getName( manID ),
					50, __FILE__, __LINE__ );
			}

			// Fill Device Description
			if( devID != UNKNOWN_ID )
			{
				fillMe->mDescription.setValue( "USB Device", 50, __FILE__, __LINE__ );
				fillMe->mModel.setValue( mUsbTable->getName( manID, devID ),
					50, __FILE__, __LINE__ );
			}
		}

		if( manID != UNKNOWN_ID || devID != UNKNOWN_ID )
		{
			ostringstream os;
			os << hex << setw( 4 ) << setfill( '0' ) << manID;
			os << hex << setw( 4 ) << setfill( '0' ) << devID;
			fillMe->mCDField.setValue( os.str( ), 90, __FILE__, __LINE__ );
		}
	}

#ifndef SIOCETHTOOL
#define SIOCETHTOOL     0x8946
#endif
	void SysFSTreeCollector::fillNetClass( Component* fillMe,
		const string& classDir )
	{
		string val = getAttrValue( classDir, "address" );
		val = val.substr( 0, 17 );
		int pos;
		while( ( pos = val.find( ':' ) ) != string::npos )
		{
			val.erase( pos, 1 );
		}

		fillMe->mNetAddr.setValue( val, 90, __FILE__, __LINE__ );

		struct ethtool_drvinfo info;
		struct ifreq ifr;
		int fd;

		fd = socket( AF_INET, SOCK_DGRAM, 0 );
		if( fd < 0 )
		{
			Logger logger;
			logger.log( "SysFsTreeCollector.fillNetClass: socket call failed.",
				LOG_WARNING );
			return;
		}

		memset( &ifr, 0, sizeof( ifr ) );
		memset( &info, 0, sizeof( ethtool_drvinfo ) );
		info.cmd = ETHTOOL_GDRVINFO;
		strncpy( ifr.ifr_name, fillMe->mAIXNames[ 0 ]->dataValue.c_str( ),
			IFNAMSIZ );
		ifr.ifr_data = (caddr_t)&info;

		if( ioctl( fd, SIOCETHTOOL, &ifr ) == -1 )
		{
			close( fd );
			Logger logger;
			logger.log( "SysFsTreeCollector.fillNetClass: ioctl call failed.",
				LOG_WARNING );
			return;
		}

		close( fd );

		val = info.fw_version;
		if( val != "" && val != "N/A" )
			fillMe->mFirmwareVersion.setValue( val, 80, __FILE__, __LINE__ );
	}

	void SysFSTreeCollector::fillSystem( System* sys )
	{
		struct utsname info;

		sys->mDescription.setValue( string( "System VPD" ), 100,
											__FILE__, __LINE__ );
		sys->mRecordType.setValue( string( "VSYS" ), 100,
											__FILE__, __LINE__ );

		if( uname( &info ) != 0 )
		{
			string message =
				"SysFSTreeCollector.fillSystem: Call to uname failed.";
			Logger logger;
			logger.log( message, LOG_WARNING );
		}
		else
		{
			ostringstream os;
			os << info.sysname << " " << info.release;
			sys->mOS.setValue( os.str( ), 100, __FILE__, __LINE__ );
			sys->mNodeName.setValue( string( info.nodename ), 50
											, __FILE__, __LINE__ );
			sys->mArch.dataValue = info.machine;
		}
	}

	void SysFSTreeCollector::fillFirmware( Component* fillMe )
	{
		Logger l;
		string msg;

		string classNode = getClassLink( fillMe );
		if (classNode.length() > 0) {
			fillMe->mFirmwareVersion.setValue( getAttrValue( classNode,
				"fw_version" ), 30, __FILE__, __LINE__ );
			fillMe->mFirmwareLevel.setValue( getAttrValue( classNode,
				"fwrev" ), 30, __FILE__, __LINE__ );
		}
	}

	string SysFSTreeCollector::resolveClassPath( const string& path )
	{
		string devPath = "/sys" + path;
		string device = devPath.substr( devPath.rfind( '/' ) + 1 );
		devPath += "/bus";
		char rel[ 1024 ] = { 0 };
		if( readlink( devPath.c_str( ), rel, 1023 ) < 0 )
		{
			return "";
		}

		ostringstream resPath;
		resPath << "/sys/bus/";
		string bus = rel;
		resPath << bus.substr( bus.rfind( '/' ) + 1 ) << "/devices/" << device;

		return resPath.str( );
	}
}
